# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.fixtures import (
    override_application_settings,
    override_generic_settings, override_ignore_status_codes)
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics
from newrelic.hooks.framework_django import django_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors

import os

import django

DJANGO_VERSION = tuple(map(int, django.get_version().split('.')[:2]))
DJANGO_SETTINGS_MODULE = os.environ.get('DJANGO_SETTINGS_MODULE', None)


def target_application():
    from _target_application import _target_application
    return _target_application


# The middleware scoped metrics are dependent on the MIDDLEWARE_CLASSES or
# MIDDLEWARE defined in the version-specific Django settings.py file.

_test_django_pre_1_10_middleware_scoped_metrics = [
    (('Function/django.middleware.common:'
            'CommonMiddleware.process_request'), 1),
    (('Function/django.contrib.sessions.middleware:'
            'SessionMiddleware.process_request'), 1),
    (('Function/django.contrib.auth.middleware:'
            'AuthenticationMiddleware.process_request'), 1),
    (('Function/django.contrib.messages.middleware:'
            'MessageMiddleware.process_request'), 1),
    (('Function/django.middleware.csrf:'
            'CsrfViewMiddleware.process_view'), 1),
    (('Function/django.contrib.messages.middleware:'
            'MessageMiddleware.process_response'), 1),
    (('Function/django.middleware.csrf:'
            'CsrfViewMiddleware.process_response'), 1),
    (('Function/django.contrib.sessions.middleware:'
            'SessionMiddleware.process_response'), 1),
    (('Function/django.middleware.common:'
            'CommonMiddleware.process_response'), 1),
    (('Function/django.middleware.gzip:'
            'GZipMiddleware.process_response'), 1),
    (('Function/newrelic.hooks.framework_django:'
            'browser_timing_insertion'), 1),
]

_test_django_post_1_10_middleware_scoped_metrics = [
    ('Function/django.middleware.security:SecurityMiddleware', 1),
    ('Function/django.contrib.sessions.middleware:SessionMiddleware', 1),
    ('Function/django.middleware.common:CommonMiddleware', 1),
    ('Function/django.middleware.csrf:CsrfViewMiddleware', 1),
    ('Function/django.contrib.auth.middleware:AuthenticationMiddleware', 1),
    ('Function/django.contrib.messages.middleware:MessageMiddleware', 1),
    ('Function/django.middleware.clickjacking:XFrameOptionsMiddleware', 1),
    ('Function/django.middleware.gzip:GZipMiddleware', 1),
]

_test_django_pre_1_10_url_resolver_scoped_metrics = [
    ('Function/django.core.urlresolvers:RegexURLResolver.resolve', 'present'),
]

_test_django_post_1_10_url_resolver_scoped_metrics = [
    ('Function/django.urls.resolvers:RegexURLResolver.resolve', 'present'),
]

_test_django_post_2_0_url_resolver_scoped_metrics = [
    ('Function/django.urls.resolvers:URLResolver.resolve', 'present'),
]

_test_application_index_scoped_metrics = [
    ('Function/django.core.handlers.wsgi:WSGIHandler.__call__', 1),
    ('Python/WSGI/Application', 1),
    ('Python/WSGI/Response', 1),
    ('Python/WSGI/Finalize', 1),
    ('Function/views:index', 1),
]

if DJANGO_VERSION >= (1, 5):
    _test_application_index_scoped_metrics.extend([
            ('Function/django.http.response:HttpResponse.close', 1)])

if DJANGO_VERSION < (1, 10):
    _test_application_index_scoped_metrics.extend(
        _test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_application_index_scoped_metrics.extend(
        _test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_application_index_scoped_metrics.extend(
        _test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == 'settings_0110_old':
    _test_application_index_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == 'settings_0110_new':
    _test_application_index_scoped_metrics.extend(
        _test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_application_index_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:index',
        scoped_metrics=_test_application_index_scoped_metrics)
@validate_code_level_metrics("views", "index")
def test_application_index():
    test_application = target_application()
    response = test_application.get('')
    response.mustcontain('INDEX RESPONSE')


@validate_transaction_metrics('views:exception')
@validate_code_level_metrics("views", "exception")
def test_application_exception():
    test_application = target_application()
    test_application.get('/exception', status=500)


_test_application_not_found_scoped_metrics = [
        ('Function/django.core.handlers.wsgi:WSGIHandler.__call__', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
]

if DJANGO_VERSION >= (1, 5):
    _test_application_not_found_scoped_metrics.extend([
            ('Function/django.http.response:HttpResponseNotFound.close', 1)])

if DJANGO_VERSION < (1, 10):
    _test_application_not_found_scoped_metrics.extend(
        _test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_application_not_found_scoped_metrics.extend(
        _test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_application_not_found_scoped_metrics.extend(
        _test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == 'settings_0110_old':
    _test_application_not_found_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)
    # The `CsrfViewMiddleware.process_view` isn't called for 404 Not Found.
    _test_application_not_found_scoped_metrics.remove(
        ('Function/django.middleware.csrf:CsrfViewMiddleware.process_view', 1))
elif DJANGO_SETTINGS_MODULE == 'settings_0110_new':
    _test_application_not_found_scoped_metrics.extend(
        _test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_application_not_found_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)
    # The `CsrfViewMiddleware.process_view` isn't called for 404 Not Found.
    _test_application_not_found_scoped_metrics.remove(
        ('Function/django.middleware.csrf:CsrfViewMiddleware.process_view', 1))


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('django.views.debug:technical_404_response',
        scoped_metrics=_test_application_not_found_scoped_metrics)
def test_application_not_found():
    test_application = target_application()
    test_application.get('/not_found', status=404)


@override_ignore_status_codes([403])
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:permission_denied')
@validate_code_level_metrics("views", "permission_denied")
def test_ignored_status_code():
    test_application = target_application()
    test_application.get('/permission_denied', status=403)


@override_ignore_status_codes([410])
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:middleware_410')
@validate_code_level_metrics("views", "middleware_410")
def test_middleware_ignore_status_codes():
    test_application = target_application()
    test_application.get('/middleware_410', status=410)


_test_application_cbv_scoped_metrics = [
        ('Function/django.core.handlers.wsgi:WSGIHandler.__call__', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/views:MyView', 1),
        ('Function/views:MyView.get', 1),
]

if DJANGO_VERSION >= (1, 5):
    _test_application_cbv_scoped_metrics.extend([
            ('Function/django.http.response:HttpResponse.close', 1)])

if DJANGO_VERSION < (1, 10):
    _test_application_cbv_scoped_metrics.extend(
        _test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_application_cbv_scoped_metrics.extend(
        _test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_application_cbv_scoped_metrics.extend(
        _test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == 'settings_0110_old':
    _test_application_cbv_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == 'settings_0110_new':
    _test_application_cbv_scoped_metrics.extend(
        _test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_application_cbv_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:MyView.get',
        scoped_metrics=_test_application_cbv_scoped_metrics)
@validate_code_level_metrics("views.MyView", "get")
def test_application_cbv():
    test_application = target_application()
    response = test_application.get('/cbv')
    response.mustcontain('CBV RESPONSE')


_test_application_deferred_cbv_scoped_metrics = [
        ('Function/django.core.handlers.wsgi:WSGIHandler.__call__', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/views:deferred_cbv', 1),
        ('Function/views:MyView.get', 1),
]

if DJANGO_VERSION >= (1, 5):
    _test_application_deferred_cbv_scoped_metrics.extend([
            ('Function/django.http.response:HttpResponse.close', 1)])

if DJANGO_VERSION < (1, 10):
    _test_application_deferred_cbv_scoped_metrics.extend(
        _test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_application_deferred_cbv_scoped_metrics.extend(
        _test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_application_deferred_cbv_scoped_metrics.extend(
        _test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == 'settings_0110_old':
    _test_application_deferred_cbv_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == 'settings_0110_new':
    _test_application_deferred_cbv_scoped_metrics.extend(
        _test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_application_deferred_cbv_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:deferred_cbv',
        scoped_metrics=_test_application_deferred_cbv_scoped_metrics)
@validate_code_level_metrics("views", "deferred_cbv")
def test_application_deferred_cbv():
    test_application = target_application()
    response = test_application.get('/deferred_cbv')
    response.mustcontain('CBV RESPONSE')


_test_html_insertion_settings = {
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': True,
    'js_agent_loader': u'<!-- NREUM HEADER -->',
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_django_middleware():
    test_application = target_application()
    response = test_application.get('/html_insertion', status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain('NREUM HEADER', 'NREUM.info')


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_django_gzip_middleware_enabled():
    test_application = target_application()

    # GZipMiddleware only fires if given the following header.

    gzip_header = {'Accept-Encoding': 'gzip'}
    response = test_application.get('/gzip_html_insertion', status=200,
            headers=gzip_header)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    # The response.text will already be gunzipped

    response.mustcontain('NREUM HEADER', 'NREUM.info')


_test_html_insertion_settings_disabled = {
    'browser_monitoring.enabled': False,
    'browser_monitoring.auto_instrument': False,
    'js_agent_loader': u'<!-- NREUM HEADER -->',
}


@override_application_settings(_test_html_insertion_settings_disabled)
def test_html_insertion_django_gzip_middleware_disabled():
    test_application = target_application()

    # GZipMiddleware only fires if given the following header.

    gzip_header = {'Accept-Encoding': 'gzip'}
    response = test_application.get('/gzip_html_insertion', status=200,
            headers=gzip_header)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    # The response.text will already be gunzipped

    response.mustcontain(no=['NREUM HEADER', 'NREUM.info'])


_test_html_insertion_manual_settings = {
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': True,
    'js_agent_loader': u'<!-- NREUM HEADER -->',
}


@override_application_settings(_test_html_insertion_manual_settings)
def test_html_insertion_manual_django_middleware():
    test_application = target_application()
    response = test_application.get('/html_insertion_manual', status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain(no=['NREUM HEADER', 'NREUM.info'])


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_unnamed_attachment_header_django_middleware():
    test_application = target_application()
    response = test_application.get(
            '/html_insertion_unnamed_attachment_header', status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain(no=['NREUM HEADER', 'NREUM.info'])


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_named_attachment_header_django_middleware():
    test_application = target_application()
    response = test_application.get(
            '/html_insertion_named_attachment_header', status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain(no=['NREUM HEADER', 'NREUM.info'])


_test_html_insertion_settings = {
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': False,
    'js_agent_loader': u'<!-- NREUM HEADER -->',
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_manual_tag_instrumentation():
    test_application = target_application()
    response = test_application.get('/template_tags')

    # Assert that the instrumentation is not inappropriately escaped

    response.mustcontain('<!-- NREUM HEADER -->',
            no=['&lt;!-- NREUM HEADER --&gt'])


_test_application_inclusion_tag_scoped_metrics = [
        ('Function/django.core.handlers.wsgi:WSGIHandler.__call__', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/views:inclusion_tag', 1),
        ('Template/Render/main.html', 1),
]

if DJANGO_VERSION < (1, 9):
    _test_application_inclusion_tag_scoped_metrics.extend([
            ('Template/Include/results.html', 1)])

if DJANGO_VERSION < (1, 10):
    _test_application_inclusion_tag_scoped_metrics.extend(
        _test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_application_inclusion_tag_scoped_metrics.extend(
        _test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_application_inclusion_tag_scoped_metrics.extend(
        _test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == 'settings_0110_old':
    _test_application_inclusion_tag_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == 'settings_0110_new':
    _test_application_inclusion_tag_scoped_metrics.extend(
        _test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_application_inclusion_tag_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)

try:
    _test_application_inclusion_tag_scoped_metrics.remove(
        (('Function/newrelic.hooks.framework_django:'
                'browser_timing_insertion'), 1)
    )
except ValueError:
    pass


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:inclusion_tag',
        scoped_metrics=_test_application_inclusion_tag_scoped_metrics)
@validate_code_level_metrics("views", "inclusion_tag")
def test_application_inclusion_tag():
    test_application = target_application()
    response = test_application.get('/inclusion_tag')
    response.mustcontain('Inclusion tag')


_test_inclusion_tag_template_tags_scoped_metrics = [
        ('Function/django.core.handlers.wsgi:WSGIHandler.__call__', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/views:inclusion_tag', 1),
        ('Template/Render/main.html', 1),
]

if DJANGO_VERSION < (1, 9):
    _test_inclusion_tag_template_tags_scoped_metrics.extend([
        ('Template/Include/results.html', 1),
        ('Template/Tag/show_results', 1)])

_test_inclusion_tag_settings = {
    'instrumentation.templates.inclusion_tag': '*'
}

if DJANGO_VERSION < (1, 10):
    _test_inclusion_tag_template_tags_scoped_metrics.extend(
        _test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_inclusion_tag_template_tags_scoped_metrics.extend(
        _test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_inclusion_tag_template_tags_scoped_metrics.extend(
        _test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == 'settings_0110_old':
    _test_inclusion_tag_template_tags_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == 'settings_0110_new':
    _test_inclusion_tag_template_tags_scoped_metrics.extend(
        _test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_inclusion_tag_template_tags_scoped_metrics.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)

try:
    _test_inclusion_tag_template_tags_scoped_metrics.remove(
        (('Function/newrelic.hooks.framework_django:'
                'browser_timing_insertion'), 1)
    )
except ValueError:
    pass


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('views:inclusion_tag',
        scoped_metrics=_test_inclusion_tag_template_tags_scoped_metrics)
@override_generic_settings(django_settings, _test_inclusion_tag_settings)
@validate_code_level_metrics("views", "inclusion_tag")
def test_inclusion_tag_template_tag_metric():
    test_application = target_application()
    response = test_application.get('/inclusion_tag')
    response.mustcontain('Inclusion tag')


_test_template_render_exception_scoped_metrics_base = [
        ('Function/django.core.handlers.wsgi:WSGIHandler.__call__', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
]

if DJANGO_VERSION < (1, 5):
    _test_template_render_exception_scoped_metrics_base.append(
            ('Function/django.http:HttpResponseServerError.close', 1))
elif DJANGO_VERSION < (1, 8):
    _test_template_render_exception_scoped_metrics_base.append(
            ('Function/django.http.response:HttpResponseServerError.close', 1))
else:
    _test_template_render_exception_scoped_metrics_base.append(
            ('Function/django.http.response:HttpResponse.close', 1))

if DJANGO_VERSION < (1, 10):
    _test_template_render_exception_scoped_metrics_base.extend(
        _test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_template_render_exception_scoped_metrics_base.extend(
        _test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_template_render_exception_scoped_metrics_base.extend(
        _test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == 'settings_0110_old':
    _test_template_render_exception_scoped_metrics_base.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == 'settings_0110_new':
    _test_template_render_exception_scoped_metrics_base.extend(
        _test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_template_render_exception_scoped_metrics_base.extend(
        _test_django_pre_1_10_middleware_scoped_metrics)

if DJANGO_VERSION < (1, 9):
    _test_template_render_exception_errors = [
        'django.template.base:TemplateSyntaxError']
else:
    _test_template_render_exception_errors = [
        'django.template.exceptions:TemplateSyntaxError']

_test_template_render_exception_function_scoped_metrics = list(
        _test_template_render_exception_scoped_metrics_base)
_test_template_render_exception_function_scoped_metrics.extend([
        ('Function/views:render_exception_function', 1),
])


@validate_transaction_errors(errors=_test_template_render_exception_errors)
@validate_transaction_metrics('views:render_exception_function',
        scoped_metrics=_test_template_render_exception_function_scoped_metrics)
@validate_code_level_metrics("views", "render_exception_function")
def test_template_render_exception_function():
    test_application = target_application()
    test_application.get('/render_exception_function', status=500)


_test_template_render_exception_class_scoped_metrics = list(
        _test_template_render_exception_scoped_metrics_base)
_test_template_render_exception_class_scoped_metrics.extend([
        ('Function/views:RenderExceptionClass', 1),
        ('Function/views:RenderExceptionClass.get', 1),
])


@validate_transaction_errors(errors=_test_template_render_exception_errors)
@validate_transaction_metrics('views:RenderExceptionClass.get',
        scoped_metrics=_test_template_render_exception_class_scoped_metrics)
@validate_code_level_metrics("views.RenderExceptionClass", "get")
def test_template_render_exception_class():
    test_application = target_application()
    test_application.get('/render_exception_class', status=500)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from graphql import (
    GraphQLArgument,
    GraphQLField,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
    GraphQLUnionType,
)

authors = [
    {
        "first_name": "New",
        "last_name": "Relic",
    },
    {
        "first_name": "Bob",
        "last_name": "Smith",
    },
    {
        "first_name": "Leslie",
        "last_name": "Jones",
    },
]

books = [
    {
        "id": 1,
        "name": "Python Agent: The Book",
        "isbn": "a-fake-isbn",
        "author": authors[0],
        "branch": "riverside",
    },
    {
        "id": 2,
        "name": "Ollies for O11y: A Sk8er's Guide to Observability",
        "isbn": "a-second-fake-isbn",
        "author": authors[1],
        "branch": "downtown",
    },
    {
        "id": 3,
        "name": "[Redacted]",
        "isbn": "a-third-fake-isbn",
        "author": authors[2],
        "branch": "riverside",
    },
]

magazines = [
    {"id": 1, "name": "Reli Updates Weekly", "issue": 1, "branch": "riverside"},
    {"id": 2, "name": "Reli Updates Weekly", "issue": 2, "branch": "downtown"},
    {"id": 3, "name": "Node Weekly", "issue": 1, "branch": "riverside"},
]


libraries = ["riverside", "downtown"]
libraries = [
    {
        "id": i + 1,
        "branch": branch,
        "magazine": [m for m in magazines if m["branch"] == branch],
        "book": [b for b in books if b["branch"] == branch],
    }
    for i, branch in enumerate(libraries)
]

storage = []


def resolve_library(parent, info, index):
    return libraries[index]


def resolve_storage_add(parent, info, string):
    storage.append(string)
    return string


def resolve_storage(parent, info):
    return storage


def resolve_search(parent, info, contains):
    search_books = [b for b in books if contains in b["name"]]
    search_magazines = [m for m in magazines if contains in m["name"]]
    return search_books + search_magazines


Author = GraphQLObjectType(
    "Author",
    {
        "first_name": GraphQLField(GraphQLString),
        "last_name": GraphQLField(GraphQLString),
    },
)

Book = GraphQLObjectType(
    "Book",
    {
        "id": GraphQLField(GraphQLInt),
        "name": GraphQLField(GraphQLString),
        "isbn": GraphQLField(GraphQLString),
        "author": GraphQLField(Author),
        "branch": GraphQLField(GraphQLString),
    },
)

Magazine = GraphQLObjectType(
    "Magazine",
    {
        "id": GraphQLField(GraphQLInt),
        "name": GraphQLField(GraphQLString),
        "issue": GraphQLField(GraphQLInt),
        "branch": GraphQLField(GraphQLString),
    },
)


Library = GraphQLObjectType(
    "Library",
    {
        "id": GraphQLField(GraphQLInt),
        "branch": GraphQLField(GraphQLString),
        "book": GraphQLField(GraphQLList(Book)),
        "magazine": GraphQLField(GraphQLList(Magazine)),
    },
)

Storage = GraphQLList(GraphQLString)


def resolve_hello(root, info):
    return "Hello!"


def resolve_echo(root, info, echo):
    return echo


def resolve_error(root, info):
    raise RuntimeError("Runtime Error!")


try:
    hello_field = GraphQLField(GraphQLString, resolver=resolve_hello)
    library_field = GraphQLField(
        Library,
        resolver=resolve_library,
        args={"index": GraphQLArgument(GraphQLNonNull(GraphQLInt))},
    )
    search_field = GraphQLField(
        GraphQLList(GraphQLUnionType("Item", (Book, Magazine), resolve_type=resolve_search)),
        args={"contains": GraphQLArgument(GraphQLNonNull(GraphQLString))},
    )
    echo_field = GraphQLField(
        GraphQLString,
        resolver=resolve_echo,
        args={"echo": GraphQLArgument(GraphQLNonNull(GraphQLString))},
    )
    storage_field = GraphQLField(
        Storage,
        resolver=resolve_storage,
    )
    storage_add_field = GraphQLField(
        Storage,
        resolver=resolve_storage_add,
        args={"string": GraphQLArgument(GraphQLNonNull(GraphQLString))},
    )
    error_field = GraphQLField(GraphQLString, resolver=resolve_error)
    error_non_null_field = GraphQLField(GraphQLNonNull(GraphQLString), resolver=resolve_error)
    error_middleware_field = GraphQLField(GraphQLString, resolver=resolve_hello)
except TypeError:
    hello_field = GraphQLField(GraphQLString, resolve=resolve_hello)
    library_field = GraphQLField(
        Library,
        resolve=resolve_library,
        args={"index": GraphQLArgument(GraphQLNonNull(GraphQLInt))},
    )
    search_field = GraphQLField(
        GraphQLList(GraphQLUnionType("Item", (Book, Magazine), resolve_type=resolve_search)),
        args={"contains": GraphQLArgument(GraphQLNonNull(GraphQLString))},
    )
    echo_field = GraphQLField(
        GraphQLString,
        resolve=resolve_echo,
        args={"echo": GraphQLArgument(GraphQLNonNull(GraphQLString))},
    )
    storage_field = GraphQLField(
        Storage,
        resolve=resolve_storage,
    )
    storage_add_field = GraphQLField(
        GraphQLString,
        resolve=resolve_storage_add,
        args={"string": GraphQLArgument(GraphQLNonNull(GraphQLString))},
    )
    error_field = GraphQLField(GraphQLString, resolve=resolve_error)
    error_non_null_field = GraphQLField(GraphQLNonNull(GraphQLString), resolve=resolve_error)
    error_middleware_field = GraphQLField(GraphQLString, resolve=resolve_hello)

query = GraphQLObjectType(
    name="Query",
    fields={
        "hello": hello_field,
        "library": library_field,
        "search": search_field,
        "echo": echo_field,
        "storage": storage_field,
        "error": error_field,
        "error_non_null": error_non_null_field,
        "error_middleware": error_middleware_field,
    },
)

mutation = GraphQLObjectType(
    name="Mutation",
    fields={
        "storage_add": storage_add_field,
    },
)

_target_application = GraphQLSchema(query=query, mutation=mutation)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import uuid

import botocore.session
import moto

from newrelic.api.background_task import background_task
from testing_support.fixtures import (
        validate_tt_segment_params, override_application_settings)
from testing_support.validators.validate_span_events import (
        validate_span_events)
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics

MOTO_VERSION = tuple(int(v) for v in moto.__version__.split('.')[:3])

# patch earlier versions of moto to support py37
if sys.version_info >= (3, 7) and MOTO_VERSION <= (1, 3, 1):
    import re
    moto.packages.responses.responses.re._pattern_type = re.Pattern


AWS_ACCESS_KEY_ID = 'AAAAAAAAAAAACCESSKEY'
AWS_SECRET_ACCESS_KEY = 'AAAAAASECRETKEY'
AWS_REGION = 'us-east-1'

TEST_TABLE = 'python-agent-test-%s' % uuid.uuid4()


_dynamodb_scoped_metrics = [
    ('Datastore/statement/DynamoDB/%s/create_table' % TEST_TABLE, 1),
    ('Datastore/statement/DynamoDB/%s/put_item' % TEST_TABLE, 1),
    ('Datastore/statement/DynamoDB/%s/get_item' % TEST_TABLE, 1),
    ('Datastore/statement/DynamoDB/%s/update_item' % TEST_TABLE, 1),
    ('Datastore/statement/DynamoDB/%s/query' % TEST_TABLE, 1),
    ('Datastore/statement/DynamoDB/%s/scan' % TEST_TABLE, 1),
    ('Datastore/statement/DynamoDB/%s/delete_item' % TEST_TABLE, 1),
    ('Datastore/statement/DynamoDB/%s/delete_table' % TEST_TABLE, 1),
]

_dynamodb_rollup_metrics = [
    ('Datastore/all', 8),
    ('Datastore/allOther', 8),
    ('Datastore/DynamoDB/all', 8),
    ('Datastore/DynamoDB/allOther', 8),
]


@override_application_settings({'distributed_tracing.enabled': True})
@validate_span_events(expected_agents=('aws.requestId',), count=8)
@validate_span_events(exact_agents={'aws.operation': 'PutItem'}, count=1)
@validate_span_events(exact_agents={'aws.operation': 'GetItem'}, count=1)
@validate_span_events(exact_agents={'aws.operation': 'DeleteItem'}, count=1)
@validate_span_events(exact_agents={'aws.operation': 'CreateTable'}, count=1)
@validate_span_events(exact_agents={'aws.operation': 'DeleteTable'}, count=1)
@validate_span_events(exact_agents={'aws.operation': 'Query'}, count=1)
@validate_span_events(exact_agents={'aws.operation': 'Scan'}, count=1)
@validate_tt_segment_params(present_params=('aws.requestId',))
@validate_transaction_metrics(
        'test_botocore_dynamodb:test_dynamodb',
        scoped_metrics=_dynamodb_scoped_metrics,
        rollup_metrics=_dynamodb_rollup_metrics,
        background_task=True)
@background_task()
@moto.mock_dynamodb2
def test_dynamodb():
    session = botocore.session.get_session()
    client = session.create_client(
            'dynamodb',
            region_name=AWS_REGION,
            aws_access_key_id=AWS_ACCESS_KEY_ID,
            aws_secret_access_key=AWS_SECRET_ACCESS_KEY
    )

    # Create table
    resp = client.create_table(
            TableName=TEST_TABLE,
            AttributeDefinitions=[
                    {'AttributeName': 'Id', 'AttributeType': 'N'},
                    {'AttributeName': 'Foo', 'AttributeType': 'S'},
            ],
            KeySchema=[
                    {'AttributeName': 'Id', 'KeyType': 'HASH'},
                    {'AttributeName': 'Foo', 'KeyType': 'RANGE'},
            ],
            ProvisionedThroughput={
                    'ReadCapacityUnits': 5,
                    'WriteCapacityUnits': 5,
            },
    )
    assert resp['TableDescription']['TableName'] == TEST_TABLE
    # moto response is ACTIVE, AWS response is CREATING
    # assert resp['TableDescription']['TableStatus'] == 'ACTIVE'

    # # AWS needs time to create the table
    # import time
    # time.sleep(15)

    # Put item
    resp = client.put_item(
            TableName=TEST_TABLE,
            Item={
                    'Id': {'N': '101'},
                    'Foo': {'S': 'hello_world'},
                    'SomeValue': {'S': 'some_random_attribute'},
            }
    )
    # No checking response, due to inconsistent return values.
    # moto returns resp['Attributes']. AWS returns resp['ResponseMetadata']

    # Get item
    resp = client.get_item(
            TableName=TEST_TABLE,
            Key={
                    'Id': {'N': '101'},
                    'Foo': {'S': 'hello_world'},
                    'SomeValue': {'S': 'some_random_attribute'},
            }
    )
    assert resp['Item']['SomeValue']['S'] == 'some_random_attribute'

    # Update item
    resp = client.update_item(
            TableName=TEST_TABLE,
            Key={
                    'Id': {'N': '101'},
                    'Foo': {'S': 'hello_world'},
                    'SomeValue': {'S': 'some_random_attribute'},
            },
            AttributeUpdates={
                    'Foo2': {
                            'Value': {'S': 'hello_world2'},
                            'Action': 'PUT'
                    },
            },
            ReturnValues='ALL_NEW',
    )
    assert resp['Attributes']['Foo2']

    # Query for item
    resp = client.query(
            TableName=TEST_TABLE,
            Select='ALL_ATTRIBUTES',
            KeyConditionExpression='#Id = :v_id',
            ExpressionAttributeNames={'#Id': 'Id'},
            ExpressionAttributeValues={':v_id': {'N': '101'}},
    )
    assert len(resp['Items']) == 1
    assert resp['Items'][0]['SomeValue']['S'] == 'some_random_attribute'

    # Scan
    resp = client.scan(TableName=TEST_TABLE)
    assert len(resp['Items']) == 1

    # Delete item
    resp = client.delete_item(
            TableName=TEST_TABLE,
            Key={
                    'Id': {'N': '101'},
                    'Foo': {'S': 'hello_world'},
            },
    )
    # No checking response, due to inconsistent return values.
    # moto returns resp['Attributes']. AWS returns resp['ResponseMetadata']

    # Delete table
    resp = client.delete_table(TableName=TEST_TABLE)
    assert resp['TableDescription']['TableName'] == TEST_TABLE
    # moto response is ACTIVE, AWS response is DELETING
    # assert resp['TableDescription']['TableStatus'] == 'DELETING'



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import uuid

import boto3
import botocore
import moto
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

MOTO_VERSION = tuple(int(v) for v in moto.__version__.split(".")[:3])

# patch earlier versions of moto to support py37
if sys.version_info >= (3, 7) and MOTO_VERSION <= (1, 3, 1):
    import re

    moto.packages.responses.responses.re._pattern_type = re.Pattern

AWS_ACCESS_KEY_ID = "AAAAAAAAAAAACCESSKEY"
AWS_SECRET_ACCESS_KEY = "AAAAAASECRETKEY"  # nosec
AWS_REGION_NAME = "us-west-2"

TEST_BUCKET = "python-agent-test-%s" % uuid.uuid4()

BOTOCORE_VERSION = tuple(map(int, botocore.__version__.split(".")))


if BOTOCORE_VERSION < (1, 7, 41):
    S3_URL = "s3-us-west-2.amazonaws.com"
    EXPECTED_BUCKET_URL = "https://%s/%s" % (S3_URL, TEST_BUCKET)
    EXPECTED_KEY_URL = EXPECTED_BUCKET_URL + "/hello_world"
elif BOTOCORE_VERSION < (1, 28):
    S3_URL = "s3.us-west-2.amazonaws.com"
    EXPECTED_BUCKET_URL = "https://%s/%s" % (S3_URL, TEST_BUCKET)
    EXPECTED_KEY_URL = EXPECTED_BUCKET_URL + "/hello_world"
else:
    S3_URL = "%s.s3.us-west-2.amazonaws.com" % TEST_BUCKET
    EXPECTED_BUCKET_URL = "https://%s/" % S3_URL
    EXPECTED_KEY_URL = EXPECTED_BUCKET_URL + "hello_world"


_s3_scoped_metrics = [
    ("External/%s/botocore/GET" % S3_URL, 2),
    ("External/%s/botocore/PUT" % S3_URL, 2),
    ("External/%s/botocore/DELETE" % S3_URL, 2),
]

_s3_rollup_metrics = [
    ("External/all", 6),
    ("External/allOther", 6),
    ("External/%s/all" % S3_URL, 6),
    ("External/%s/botocore/GET" % S3_URL, 2),
    ("External/%s/botocore/PUT" % S3_URL, 2),
    ("External/%s/botocore/DELETE" % S3_URL, 2),
]


@override_application_settings({"distributed_tracing.enabled": True})
@validate_span_events(exact_agents={"aws.operation": "CreateBucket"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "PutObject"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "ListObjects"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "GetObject"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "DeleteObject"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "DeleteBucket"}, count=1)
@validate_span_events(exact_agents={"http.url": EXPECTED_BUCKET_URL}, count=3)
@validate_span_events(exact_agents={"http.url": EXPECTED_KEY_URL}, count=3)
@validate_transaction_metrics(
    "test_boto3_s3:test_s3", scoped_metrics=_s3_scoped_metrics, rollup_metrics=_s3_rollup_metrics, background_task=True
)
@background_task()
@moto.mock_s3
def test_s3():
    client = boto3.client(
        "s3",
        aws_access_key_id=AWS_ACCESS_KEY_ID,
        aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
        region_name=AWS_REGION_NAME,
    )

    # Create bucket
    resp = client.create_bucket(Bucket=TEST_BUCKET, CreateBucketConfiguration={"LocationConstraint": AWS_REGION_NAME})
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200

    # Put object
    resp = client.put_object(Bucket=TEST_BUCKET, Key="hello_world", Body=b"hello_world_content")
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200

    # List bucket
    resp = client.list_objects(Bucket=TEST_BUCKET)
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert len(resp["Contents"]) == 1
    assert resp["Contents"][0]["Key"] == "hello_world"

    # Get object
    resp = client.get_object(Bucket=TEST_BUCKET, Key="hello_world")
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert resp["Body"].read() == b"hello_world_content"

    # Delete object
    resp = client.delete_object(Bucket=TEST_BUCKET, Key="hello_world")
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 204

    # Delete bucket
    resp = client.delete_bucket(Bucket=TEST_BUCKET)
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 204



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.background_task import background_task
from testing_support.fixtures import reset_core_stats_engine
from testing_support.validators.validate_custom_metrics_outside_transaction import validate_custom_metrics_outside_transaction
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics


def exercise_logging(logger):
    logger.warning("C")
    logger.error("D")
    logger.critical("E")
    
    assert len(logger.caplog.records) == 3


_test_logging_unscoped_metrics = [
    ("Logging/lines", 3),
    ("Logging/lines/WARNING", 1),
    ("Logging/lines/ERROR", 1),
    ("Logging/lines/CRITICAL", 1),
]

@reset_core_stats_engine()
def test_logging_metrics_inside_transaction(logger):
    @validate_transaction_metrics(
        "test_metrics:test_logging_metrics_inside_transaction.<locals>.test",
        custom_metrics=_test_logging_unscoped_metrics,
        background_task=True,
    )
    @background_task()
    def test():
        exercise_logging(logger)

    test()


@reset_core_stats_engine()
def test_logging_metrics_outside_transaction(logger):
    @validate_custom_metrics_outside_transaction(_test_logging_unscoped_metrics)
    @reset_core_stats_engine()
    def test():
        exercise_logging(logger)

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from elasticsearch import Elasticsearch

try:
    from elastic_transport import RoundRobinSelector
except ImportError:
    from elasticsearch.connection_pool import RoundRobinSelector

from conftest import ES_MULTIPLE_SETTINGS, ES_VERSION
from testing_support.fixtures import override_application_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

# Settings

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
}

# Metrics

_base_scoped_metrics = (("Datastore/statement/Elasticsearch/contacts/index", 2),)

_base_rollup_metrics = (
    ("Datastore/all", 3),
    ("Datastore/allOther", 3),
    ("Datastore/Elasticsearch/all", 3),
    ("Datastore/Elasticsearch/allOther", 3),
    ("Datastore/operation/Elasticsearch/index", 2),
    ("Datastore/operation/Elasticsearch/mget", 1),
    ("Datastore/statement/Elasticsearch/contacts/index", 2),
)

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

if len(ES_MULTIPLE_SETTINGS) > 1:
    es_1 = ES_MULTIPLE_SETTINGS[0]
    es_2 = ES_MULTIPLE_SETTINGS[1]

    host_1 = instance_hostname(es_1["host"])
    port_1 = es_1["port"]

    host_2 = instance_hostname(es_2["host"])
    port_2 = es_2["port"]

    instance_metric_name_1 = "Datastore/instance/Elasticsearch/%s/%s" % (host_1, port_1)
    instance_metric_name_2 = "Datastore/instance/Elasticsearch/%s/%s" % (host_2, port_2)

    _enable_rollup_metrics.extend(
        [
            (instance_metric_name_1, 2),
            (instance_metric_name_2, 1),
        ]
    )

    _disable_rollup_metrics.extend(
        [
            (instance_metric_name_1, None),
            (instance_metric_name_2, None),
        ]
    )


@pytest.fixture(scope="module")
def client():
    urls = ["http://%s:%s" % (db["host"], db["port"]) for db in ES_MULTIPLE_SETTINGS]
    # When selecting a connection from the pool, use the round robin method.
    # This is actually the default already. Using round robin will ensure that
    # doing two db calls will mean elastic search is talking to two different
    # dbs.
    if ES_VERSION >= (8,):
        client = Elasticsearch(urls, node_selector_class=RoundRobinSelector, randomize_hosts=False)
    else:
        client = Elasticsearch(urls, selector_class=RoundRobinSelector, randomize_hosts=False)
    return client


# Query


def _exercise_es_multi(es):
    # set on db 1
    if ES_VERSION >= (8,):
        es.index(index="contacts", body={"name": "Joe Tester", "age": 25, "title": "QA Engineer"}, id=1)
        # set on db 2
        es.index(index="contacts", body={"name": "Jane Tester", "age": 22, "title": "Senior QA Engineer"}, id=2)
    else:
        es.index(
            index="contacts", doc_type="person", body={"name": "Joe Tester", "age": 25, "title": "QA Engineer"}, id=1
        )
        # set on db 2
        es.index(
            index="contacts",
            doc_type="person",
            body={"name": "Jane Tester", "age": 22, "title": "Senior QA Engineer"},
            id=2,
        )

    # ask db 1, will return info from db 1 and 2
    mget_body = {
        "docs": [
            {"_id": 1, "_index": "contacts"},
            {"_id": 2, "_index": "contacts"},
        ]
    }

    results = es.mget(body=mget_body)
    assert len(results["docs"]) == 2


# Test


@pytest.mark.skipif(len(ES_MULTIPLE_SETTINGS) < 2, reason="Test environment not configured with multiple databases.")
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_mget:test_multi_get_enabled",
    scoped_metrics=_enable_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_multi_get_enabled(client):
    _exercise_es_multi(client)


@pytest.mark.skipif(len(ES_MULTIPLE_SETTINGS) < 2, reason="Test environment not configured with multiple databases.")
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_mget:test_multi_get_disabled",
    scoped_metrics=_disable_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_multi_get_disabled(client):
    _exercise_es_multi(client)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from ariadne import (
    MutationType,
    QueryType,
    UnionType,
    load_schema_from_path,
    make_executable_schema,
)
from ariadne.asgi import GraphQL as GraphQLASGI
from ariadne.wsgi import GraphQL as GraphQLWSGI

schema_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), "schema.graphql")
type_defs = load_schema_from_path(schema_file)


authors = [
    {
        "first_name": "New",
        "last_name": "Relic",
    },
    {
        "first_name": "Bob",
        "last_name": "Smith",
    },
    {
        "first_name": "Leslie",
        "last_name": "Jones",
    },
]

books = [
    {
        "id": 1,
        "name": "Python Agent: The Book",
        "isbn": "a-fake-isbn",
        "author": authors[0],
        "branch": "riverside",
    },
    {
        "id": 2,
        "name": "Ollies for O11y: A Sk8er's Guide to Observability",
        "isbn": "a-second-fake-isbn",
        "author": authors[1],
        "branch": "downtown",
    },
    {
        "id": 3,
        "name": "[Redacted]",
        "isbn": "a-third-fake-isbn",
        "author": authors[2],
        "branch": "riverside",
    },
]

magazines = [
    {"id": 1, "name": "Reli Updates Weekly", "issue": 1, "branch": "riverside"},
    {"id": 2, "name": "Reli Updates Weekly", "issue": 2, "branch": "downtown"},
    {"id": 3, "name": "Node Weekly", "issue": 1, "branch": "riverside"},
]


libraries = ["riverside", "downtown"]
libraries = [
    {
        "id": i + 1,
        "branch": branch,
        "magazine": [m for m in magazines if m["branch"] == branch],
        "book": [b for b in books if b["branch"] == branch],
    }
    for i, branch in enumerate(libraries)
]

storage = []


mutation = MutationType()


@mutation.field("storage_add")
def mutate(self, info, string):
    storage.append(string)
    return {"string": string}


item = UnionType("Item")


@item.type_resolver
def resolve_type(obj, *args):
    if "isbn" in obj:
        return "Book"
    elif "issue" in obj:  # pylint: disable=R1705
        return "Magazine"

    return None


query = QueryType()


@query.field("library")
def resolve_library(self, info, index):
    return libraries[index]


@query.field("storage")
def resolve_storage(self, info):
    return storage


@query.field("search")
def resolve_search(self, info, contains):
    search_books = [b for b in books if contains in b["name"]]
    search_magazines = [m for m in magazines if contains in m["name"]]
    return search_books + search_magazines


@query.field("hello")
def resolve_hello(self, info):
    return "Hello!"


@query.field("echo")
def resolve_echo(self, info, echo):
    return echo


@query.field("error_non_null")
@query.field("error")
def resolve_error(self, info):
    raise RuntimeError("Runtime Error!")


_target_application = make_executable_schema(type_defs, query, mutation, item)
_target_asgi_application = GraphQLASGI(_target_application)
_target_wsgi_application = GraphQLWSGI(_target_application)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest

from testing_support.validators.validate_apdex_metrics import (
        validate_apdex_metrics)
from testing_support.sample_applications import simple_app


normal_application = webtest.TestApp(simple_app)


# NOTE: This test validates that the server-side apdex_t is set to 0.5
# If the server-side configuration changes, this test will start to fail.


@validate_apdex_metrics(
    name='',
    group='Uri',
    apdex_t_min=0.5,
    apdex_t_max=0.5,
)
def test_apdex():
    normal_application.get('/')



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import gc
import sys
import time

import pytest
from testing_support.fixtures import capture_transaction_metrics, validate_tt_parenting
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.api.database_trace import database_trace
from newrelic.api.datastore_trace import datastore_trace
from newrelic.api.external_trace import external_trace
from newrelic.api.function_trace import function_trace
from newrelic.api.memcache_trace import memcache_trace
from newrelic.api.message_trace import message_trace

is_pypy = hasattr(sys, "pypy_version_info")
asyncio = pytest.importorskip("asyncio")


@pytest.mark.parametrize(
    "trace,metric",
    [
        (functools.partial(function_trace, name="simple_gen"), "Function/simple_gen"),
        (functools.partial(external_trace, library="lib", url="http://foo.com"), "External/foo.com/lib/"),
        (functools.partial(database_trace, "select * from foo"), "Datastore/statement/None/foo/select"),
        (functools.partial(datastore_trace, "lib", "foo", "bar"), "Datastore/statement/lib/foo/bar"),
        (functools.partial(message_trace, "lib", "op", "typ", "name"), "MessageBroker/lib/typ/op/Named/name"),
        (functools.partial(memcache_trace, "cmd"), "Memcache/cmd"),
    ],
)
def test_coroutine_timing(trace, metric):
    @trace()
    def simple_gen():
        time.sleep(0.1)
        yield
        time.sleep(0.1)

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @validate_transaction_metrics(
        "test_coroutine_timing", background_task=True, scoped_metrics=[(metric, 1)], rollup_metrics=[(metric, 1)]
    )
    @background_task(name="test_coroutine_timing")
    def _test():
        for _ in simple_gen():
            pass

    _test()

    # Check that coroutines time the total call time (including pauses)
    metric_key = (metric, "")
    assert full_metrics[metric_key].total_call_time >= 0.2


@validate_tt_parenting(
    (
        "TransactionNode",
        [
            (
                "FunctionNode",
                [
                    ("FunctionNode", []),
                    ("FunctionNode", []),
                ],
            ),
        ],
    )
)
@validate_transaction_metrics(
    "test_coroutine_siblings",
    background_task=True,
    scoped_metrics=[("Function/child", 2)],
    rollup_metrics=[("Function/child", 2)],
)
@background_task(name="test_coroutine_siblings")
def test_coroutine_siblings(event_loop):
    # The expected segment map looks like this
    # parent
    # | child
    # | child

    # This test checks for the presence of 2 child metrics (which wouldn't be
    # the case if child was a child of child since child is terminal)

    @function_trace("child", terminal=True)
    async def child(wait, event=None):
        if event:
            event.set()
        await wait.wait()

    async def middle():
        wait = asyncio.Event()
        started = asyncio.Event()

        child_0 = asyncio.ensure_future(child(wait, started))

        # Wait for the first child to start
        await started.wait()

        child_1 = asyncio.ensure_future(child(wait))

        # Allow children to complete
        wait.set()
        await child_1
        await child_0

    @function_trace("parent")
    async def parent():
        await asyncio.ensure_future(middle())

    event_loop.run_until_complete(parent())


class MyException(Exception):
    pass


@validate_transaction_metrics(
    "test_coroutine_error",
    background_task=True,
    scoped_metrics=[("Function/coro", 1)],
    rollup_metrics=[("Function/coro", 1)],
)
@validate_transaction_errors(errors=["test_coroutine_trace:MyException"])
def test_coroutine_error():
    @function_trace(name="coro")
    def coro():
        yield

    @background_task(name="test_coroutine_error")
    def _test():
        gen = coro()
        gen.send(None)
        gen.throw(MyException)

    with pytest.raises(MyException):
        _test()


@validate_transaction_metrics(
    "test_coroutine_caught_exception",
    background_task=True,
    scoped_metrics=[("Function/coro", 1)],
    rollup_metrics=[("Function/coro", 1)],
)
@validate_transaction_errors(errors=[])
def test_coroutine_caught_exception():
    @function_trace(name="coro")
    def coro():
        for _ in range(2):
            time.sleep(0.1)
            try:
                yield
            except ValueError:
                pass

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @background_task(name="test_coroutine_caught_exception")
    def _test():
        gen = coro()
        # kickstart the generator (the try/except logic is inside the
        # generator)
        next(gen)
        gen.throw(ValueError)

        # consume the generator
        for _ in gen:
            pass

    # The ValueError should not be reraised
    _test()

    assert full_metrics[("Function/coro", "")].total_call_time >= 0.2


@validate_transaction_metrics(
    "test_coroutine_handles_terminal_nodes",
    background_task=True,
    scoped_metrics=[("Function/parent", 1), ("Function/coro", None)],
    rollup_metrics=[("Function/parent", 1), ("Function/coro", None)],
)
def test_coroutine_handles_terminal_nodes():
    # sometimes coroutines can be called underneath terminal nodes
    # In this case, the trace shouldn't actually be created and we also
    # shouldn't get any errors

    @function_trace(name="coro")
    def coro():
        yield
        time.sleep(0.1)

    @function_trace(name="parent", terminal=True)
    def parent():
        # parent calls child
        for _ in coro():
            pass

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @background_task(name="test_coroutine_handles_terminal_nodes")
    def _test():
        parent()

    _test()

    metric_key = ("Function/parent", "")
    assert full_metrics[metric_key].total_exclusive_call_time >= 0.1


@validate_transaction_metrics(
    "test_coroutine_close_ends_trace",
    background_task=True,
    scoped_metrics=[("Function/coro", 1)],
    rollup_metrics=[("Function/coro", 1)],
)
@background_task(name="test_coroutine_close_ends_trace")
def test_coroutine_close_ends_trace():
    @function_trace(name="coro")
    def coro():
        yield

    gen = coro()

    # kickstart the coroutine
    next(gen)

    # trace should be ended/recorded by close
    gen.close()

    # We may call gen.close as many times as we want
    gen.close()


@validate_tt_parenting(
    (
        "TransactionNode",
        [
            (
                "FunctionNode",
                [
                    ("FunctionNode", []),
                ],
            ),
        ],
    )
)
@validate_transaction_metrics(
    "test_coroutine_parents",
    background_task=True,
    scoped_metrics=[("Function/child", 1), ("Function/parent", 1)],
    rollup_metrics=[("Function/child", 1), ("Function/parent", 1)],
)
def test_coroutine_parents():
    @function_trace(name="child")
    def child():
        yield
        time.sleep(0.1)
        yield

    @function_trace(name="parent")
    def parent():
        time.sleep(0.1)
        yield
        for _ in child():
            pass

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @background_task(name="test_coroutine_parents")
    def _test():
        for _ in parent():
            pass

    _test()

    # Check that the child time is subtracted from the parent time (parenting
    # relationship is correctly established)
    key = ("Function/parent", "")
    assert full_metrics[key].total_exclusive_call_time < 0.2


@validate_transaction_metrics(
    "test_throw_yields_a_value",
    background_task=True,
    scoped_metrics=[("Function/coro", 1)],
    rollup_metrics=[("Function/coro", 1)],
)
@background_task(name="test_throw_yields_a_value")
def test_throw_yields_a_value():
    @function_trace(name="coro")
    def coro():
        for _ in range(2):
            try:
                yield
            except MyException:
                yield "foobar"

    c = coro()

    # kickstart the coroutine
    next(c)

    assert c.throw(MyException) == "foobar"

    # finish consumption of the coroutine if necessary
    for _ in c:
        pass


@pytest.mark.parametrize(
    "trace",
    [
        function_trace(name="simple_gen"),
        external_trace(library="lib", url="http://foo.com"),
        database_trace("select * from foo"),
        datastore_trace("lib", "foo", "bar"),
        message_trace("lib", "op", "typ", "name"),
        memcache_trace("cmd"),
    ],
)
def test_coroutine_functions_outside_of_transaction(trace):
    @trace
    def coro():
        for _ in range(2):
            yield "foo"

    assert [_ for _ in coro()] == ["foo", "foo"]


@validate_transaction_metrics(
    "test_catching_generator_exit_causes_runtime_error",
    background_task=True,
    scoped_metrics=[("Function/coro", 1)],
    rollup_metrics=[("Function/coro", 1)],
)
@background_task(name="test_catching_generator_exit_causes_runtime_error")
def test_catching_generator_exit_causes_runtime_error():
    @function_trace(name="coro")
    def coro():
        try:
            yield
        except GeneratorExit:
            yield

    gen = coro()

    # kickstart the coroutine (we're inside the try now)
    gen.send(None)

    # Generators cannot catch generator exit exceptions (which are injected by
    # close). This will result in a runtime error.
    with pytest.raises(RuntimeError):
        gen.close()

    if is_pypy:
        gen = None
        gc.collect()


@validate_transaction_metrics(
    "test_coroutine_time_excludes_creation_time",
    background_task=True,
    scoped_metrics=[("Function/coro", 1)],
    rollup_metrics=[("Function/coro", 1)],
)
def test_coroutine_time_excludes_creation_time():
    @function_trace(name="coro")
    def coro():
        yield

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @background_task(name="test_coroutine_time_excludes_creation_time")
    def _test():
        gen = coro()
        time.sleep(0.1)
        for _ in gen:
            pass

    _test()

    # check that the trace does not include the time between creation and
    # consumption
    assert full_metrics[("Function/coro", "")].total_call_time < 0.1


@pytest.mark.parametrize("nr_transaction", [True, False])
def test_incomplete_coroutine(nr_transaction):
    @function_trace(name="coro")
    def coro():
        for _ in range(5):
            yield

    def _test():
        c = coro()

        for _ in c:
            break

        if is_pypy:
            # pypy is not guaranteed to delete the coroutine when it goes out
            # of scope. This code "helps" pypy along. The test above is really
            # just to verify that incomplete coroutines will "eventually" be
            # cleaned up. In pypy, unfortunately that means it may not be
            # reported all the time. A customer would be expected to call gc
            # directly; however, they already have to handle this case since
            # incomplete generators are well documented as having problems with
            # pypy's gc.

            # See:
            # http://doc.pypy.org/en/latest/cpython_differences.html#differences-related-to-garbage-collection-strategies
            # https://bitbucket.org/pypy/pypy/issues/736
            del c
            import gc

            gc.collect()

    if nr_transaction:
        _test = validate_transaction_metrics(
            "test_incomplete_coroutine",
            background_task=True,
            scoped_metrics=[("Function/coro", 1)],
            rollup_metrics=[("Function/coro", 1)],
        )(background_task(name="test_incomplete_coroutine")(_test))

    _test()


def test_trace_outlives_transaction(event_loop):
    task = []
    running, finish = asyncio.Event(), asyncio.Event()

    @function_trace(name="coro")
    async def _coro():
        running.set()
        await finish.wait()

    async def parent():
        task.append(asyncio.ensure_future(_coro()))
        await running.wait()

    @validate_transaction_metrics(
        "test_trace_outlives_transaction",
        background_task=True,
        scoped_metrics=(("Function/coro", None),),
    )
    @background_task(name="test_trace_outlives_transaction")
    def _test():
        event_loop.run_until_complete(parent())

    _test()
    finish.set()
    event_loop.run_until_complete(task.pop())


if sys.version_info >= (3, 5):
    from _test_async_coroutine_trace import *  # NOQA



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import pytest
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.api.database_trace import DatabaseTrace
from newrelic.api.datastore_trace import DatastoreTrace
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.function_trace import FunctionTrace
from newrelic.api.graphql_trace import GraphQLOperationTrace, GraphQLResolverTrace
from newrelic.api.memcache_trace import MemcacheTrace
from newrelic.api.message_trace import MessageTrace
from newrelic.api.solr_trace import SolrTrace
from newrelic.api.transaction import current_transaction, end_of_transaction


@validate_transaction_metrics(
    "test_trace_after_end_of_transaction",
    background_task=True,
    scoped_metrics=[("Function/foobar", None)],
)
@background_task(name="test_trace_after_end_of_transaction")
def test_trace_after_end_of_transaction(caplog):
    end_of_transaction()
    with FunctionTrace("foobar"):
        pass

    error_messages = [record for record in caplog.records if record.levelno >= logging.ERROR]
    assert not error_messages


@pytest.mark.parametrize(
    "trace_type,args",
    (
        (DatabaseTrace, ("select * from foo",)),
        (DatastoreTrace, ("db_product", "db_target", "db_operation")),
        (ExternalTrace, ("lib", "url")),
        (FunctionTrace, ("name",)),
        (GraphQLOperationTrace, ()),
        (GraphQLResolverTrace, ()),
        (MemcacheTrace, ("command",)),
        (MessageTrace, ("lib", "operation", "dst_type", "dst_name")),
        (SolrTrace, ("lib", "command")),
    ),
)
@background_task()
def test_trace_finalizes_with_transaction_missing_settings(monkeypatch, trace_type, args):
    txn = current_transaction()
    try:
        with trace_type(*args):
            # Validate no errors are raised when finalizing trace with no settings
            monkeypatch.setattr(txn, "_settings", None)
    finally:
        # Ensure transaction still has settings when it exits to prevent other crashes making errors hard to read
        monkeypatch.undo()
        assert txn.settings



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import pytest
from testing_support.asgi_testing import AsgiTest
from testing_support.fixtures import dt_enabled
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_span_events import validate_span_events


@pytest.fixture(scope="session")
def graphql_asgi_run():
    """Wrapper function to simulate framework_graphql test behavior."""
    from _target_application import _target_asgi_application

    app = AsgiTest(_target_asgi_application)

    def execute(query):
        return app.make_request(
            "POST",
            "/",
            headers={"Content-Type": "application/json"},
            body=json.dumps({"query": query}),
        )

    return execute


@dt_enabled
def test_query_and_mutation_asgi(graphql_asgi_run):
    from graphql import __version__ as version

    from newrelic.hooks.framework_strawberry import framework_details

    FRAMEWORK_METRICS = [
        ("Python/Framework/Strawberry/%s" % framework_details()[1], 1),
        ("Python/Framework/GraphQL/%s" % version, 1),
    ]
    _test_mutation_scoped_metrics = [
        ("GraphQL/resolve/Strawberry/storage_add", 1),
        ("GraphQL/operation/Strawberry/mutation/<anonymous>/storage_add", 1),
    ]
    _test_query_scoped_metrics = [
        ("GraphQL/resolve/Strawberry/storage", 1),
        ("GraphQL/operation/Strawberry/query/<anonymous>/storage", 1),
    ]
    _test_unscoped_metrics = [
        ("WebTransaction", 1),
        ("GraphQL/all", 1),
        ("GraphQL/Strawberry/all", 1),
        ("GraphQL/allWeb", 1),
        ("GraphQL/Strawberry/allWeb", 1),
    ]
    _test_mutation_unscoped_metrics = _test_unscoped_metrics + _test_mutation_scoped_metrics
    _test_query_unscoped_metrics = _test_unscoped_metrics + _test_query_scoped_metrics

    _expected_mutation_operation_attributes = {
        "graphql.operation.type": "mutation",
        "graphql.operation.name": "<anonymous>",
    }
    _expected_mutation_resolver_attributes = {
        "graphql.field.name": "storage_add",
        "graphql.field.parentType": "Mutation",
        "graphql.field.path": "storage_add",
        "graphql.field.returnType": "String!",
    }
    _expected_query_operation_attributes = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "<anonymous>",
    }
    _expected_query_resolver_attributes = {
        "graphql.field.name": "storage",
        "graphql.field.parentType": "Query",
        "graphql.field.path": "storage",
        "graphql.field.returnType": "[String!]!",
    }

    @validate_transaction_metrics(
        "query/<anonymous>/storage",
        "GraphQL",
        scoped_metrics=_test_query_scoped_metrics,
        rollup_metrics=_test_query_unscoped_metrics + FRAMEWORK_METRICS,
    )
    @validate_transaction_metrics(
        "mutation/<anonymous>/storage_add",
        "GraphQL",
        scoped_metrics=_test_mutation_scoped_metrics,
        rollup_metrics=_test_mutation_unscoped_metrics + FRAMEWORK_METRICS,
        index=-2,
    )
    @validate_span_events(exact_agents=_expected_mutation_operation_attributes, index=-2)
    @validate_span_events(exact_agents=_expected_mutation_resolver_attributes, index=-2)
    @validate_span_events(exact_agents=_expected_query_operation_attributes)
    @validate_span_events(exact_agents=_expected_query_resolver_attributes)
    def _test():
        response = graphql_asgi_run('mutation { storage_add(string: "abc") }')
        assert response.status == 200
        response = json.loads(response.body.decode("utf-8"))
        assert not response.get("errors")

        response = graphql_asgi_run("query { storage }")
        assert response.status == 200
        response = json.loads(response.body.decode("utf-8"))
        assert not response.get("errors")

        # These are separate assertions because pypy stores 'abc' as a unicode string while other Python versions do not
        assert "storage" in str(response.get("data"))
        assert "abc" in str(response.get("data"))

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3 as database
import os
import sys

is_pypy = hasattr(sys, 'pypy_version_info')

from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_database_trace_inputs import validate_database_trace_inputs

from newrelic.api.background_task import background_task

DATABASE_DIR = os.environ.get('TOX_ENVDIR', '.')
DATABASE_NAME = ':memory:'

_test_execute_via_cursor_scoped_metrics = [
        ('Function/_sqlite3:connect', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/select', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/insert', 2),
        ('Datastore/statement/SQLite/datastore_sqlite/update', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/delete', 1),
        ('Datastore/operation/SQLite/drop', 1),
        ('Datastore/operation/SQLite/create', 1),
        ('Datastore/operation/SQLite/commit', 3),
        ('Datastore/operation/SQLite/rollback', 1)]

_test_execute_via_cursor_rollup_metrics = [
        ('Datastore/all', 12),
        ('Datastore/allOther', 12),
        ('Datastore/SQLite/all', 12),
        ('Datastore/SQLite/allOther', 12),
        ('Datastore/operation/SQLite/select', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/select', 1),
        ('Datastore/operation/SQLite/insert', 2),
        ('Datastore/statement/SQLite/datastore_sqlite/insert', 2),
        ('Datastore/operation/SQLite/update', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/update', 1),
        ('Datastore/operation/SQLite/delete', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/delete', 1),
        ('Datastore/operation/SQLite/drop', 1),
        ('Datastore/operation/SQLite/create', 1),
        ('Datastore/operation/SQLite/commit', 3),
        ('Datastore/operation/SQLite/rollback', 1)]

if is_pypy:
    _test_execute_via_cursor_scoped_metrics.extend([
        ('Function/_sqlite3:Connection.__exit__', 1)])
    _test_execute_via_cursor_rollup_metrics.extend([
        ('Function/_sqlite3:Connection.__exit__', 1)])
else:
    _test_execute_via_cursor_scoped_metrics.extend([
        ('Function/sqlite3:Connection.__exit__', 1)])
    _test_execute_via_cursor_rollup_metrics.extend([
        ('Function/sqlite3:Connection.__exit__', 1)])

@validate_transaction_metrics('test_database:test_execute_via_cursor',
        scoped_metrics=_test_execute_via_cursor_scoped_metrics,
        rollup_metrics=_test_execute_via_cursor_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor():
    with database.connect(DATABASE_NAME) as connection:
        cursor = connection.cursor()

        cursor.execute("""drop table if exists datastore_sqlite""")

        cursor.execute("""create table datastore_sqlite (a, b, c)""")

        cursor.executemany("""insert into datastore_sqlite values (?, ?, ?)""",
                [(1, 1.0, '1.0'), (2, 2.2, '2.2'), (3, 3.3, '3.3')])

        test_data = [(4, 4.0, '4.0'), (5, 5.5, '5.5'), (6, 6.6, '6.6')]

        cursor.executemany("""insert into datastore_sqlite values (?, ?, ?)""",
                           ((value) for value in test_data))

        cursor.execute("""select * from datastore_sqlite""")

        assert len(list(cursor)) == 6

        cursor.execute("""update datastore_sqlite set a=?, b=?, """
                """c=? where a=?""", (4, 4.0, '4.0', 1))

        script = """delete from datastore_sqlite where a = 2;"""
        cursor.executescript(script)

        connection.commit()
        connection.rollback()
        connection.commit()

_test_execute_via_connection_scoped_metrics = [
        ('Function/_sqlite3:connect', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/select', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/insert', 2),
        ('Datastore/statement/SQLite/datastore_sqlite/update', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/delete', 1),
        ('Datastore/operation/SQLite/drop', 1),
        ('Datastore/operation/SQLite/create', 1),
        ('Datastore/operation/SQLite/commit', 3),
        ('Datastore/operation/SQLite/rollback', 1)]

_test_execute_via_connection_rollup_metrics = [
        ('Datastore/all', 12),
        ('Datastore/allOther', 12),
        ('Datastore/SQLite/all', 12),
        ('Datastore/SQLite/allOther', 12),
        ('Datastore/operation/SQLite/select', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/select', 1),
        ('Datastore/operation/SQLite/insert', 2),
        ('Datastore/statement/SQLite/datastore_sqlite/insert', 2),
        ('Datastore/operation/SQLite/update', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/update', 1),
        ('Datastore/operation/SQLite/delete', 1),
        ('Datastore/statement/SQLite/datastore_sqlite/delete', 1),
        ('Datastore/operation/SQLite/drop', 1),
        ('Datastore/operation/SQLite/create', 1),
        ('Datastore/operation/SQLite/commit', 3),
        ('Datastore/operation/SQLite/rollback', 1)]

if is_pypy:
    _test_execute_via_connection_scoped_metrics.extend([
        ('Function/_sqlite3:Connection.__enter__', 1),
        ('Function/_sqlite3:Connection.__exit__', 1)])
    _test_execute_via_connection_rollup_metrics.extend([
        ('Function/_sqlite3:Connection.__enter__', 1),
        ('Function/_sqlite3:Connection.__exit__', 1)])
else:
    _test_execute_via_connection_scoped_metrics.extend([
        ('Function/sqlite3:Connection.__enter__', 1),
        ('Function/sqlite3:Connection.__exit__', 1)])
    _test_execute_via_connection_rollup_metrics.extend([
        ('Function/sqlite3:Connection.__enter__', 1),
        ('Function/sqlite3:Connection.__exit__', 1)])

@validate_transaction_metrics('test_database:test_execute_via_connection',
        scoped_metrics=_test_execute_via_connection_scoped_metrics,
        rollup_metrics=_test_execute_via_connection_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_connection():
    with database.connect(DATABASE_NAME) as connection:
        connection.execute("""drop table if exists datastore_sqlite""")

        connection.execute("""create table datastore_sqlite (a, b, c)""")

        connection.executemany("""insert into datastore_sqlite values """
                """(?, ?, ?)""", [(1, 1.0, '1.0'), (2, 2.2, '2.2'),
                (3, 3.3, '3.3')])

        test_data = [(4, 4.0, '4.0'), (5, 5.5, '5.5'), (6, 6.6, '6.6')]

        connection.executemany("""insert into datastore_sqlite values (?, ?, ?)""",
                           ((value) for value in test_data))

        cursor = connection.execute("""select * from datastore_sqlite""")

        assert len(list(cursor)) == 6

        connection.execute("""update datastore_sqlite set a=?, b=?, """
                """c=? where a=?""", (4, 4.0, '4.0', 1))

        script = """delete from datastore_sqlite where a = 2;"""
        connection.executescript(script)

        connection.commit()
        connection.rollback()
        connection.commit()

_test_rollback_on_exception_scoped_metrics = [
        ('Function/_sqlite3:connect', 1),
        ('Datastore/operation/SQLite/rollback', 1)]

_test_rollback_on_exception_rollup_metrics = [
        ('Datastore/all', 2),
        ('Datastore/allOther', 2),
        ('Datastore/SQLite/all', 2),
        ('Datastore/SQLite/allOther', 2),
        ('Datastore/operation/SQLite/rollback', 1)]

if is_pypy:
    _test_rollback_on_exception_scoped_metrics.extend([
        ('Function/_sqlite3:Connection.__enter__', 1),
        ('Function/_sqlite3:Connection.__exit__', 1)])
    _test_rollback_on_exception_rollup_metrics.extend([
        ('Function/_sqlite3:Connection.__enter__', 1),
        ('Function/_sqlite3:Connection.__exit__', 1)])
else:
    _test_rollback_on_exception_scoped_metrics.extend([
        ('Function/sqlite3:Connection.__enter__', 1),
        ('Function/sqlite3:Connection.__exit__', 1)])
    _test_rollback_on_exception_rollup_metrics.extend([
        ('Function/sqlite3:Connection.__enter__', 1),
        ('Function/sqlite3:Connection.__exit__', 1)])

@validate_transaction_metrics('test_database:test_rollback_on_exception',
        scoped_metrics=_test_rollback_on_exception_scoped_metrics,
        rollup_metrics=_test_rollback_on_exception_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_rollback_on_exception():
    try:
        with database.connect(DATABASE_NAME) as connection:
            raise RuntimeError('error')
    except RuntimeError:
        pass
